// lawnmower-man@mail.ru

#include<vector>
#include<assert.h>
#include<string.h>
#include<algorithm>
#include<stdio.h>


typedef unsigned int uint;
typedef unsigned char byte;
typedef std::vector<uint> IntArr;
typedef std::vector<byte> ByteArr;

template<typename T = byte>
class SuffArr {
	IntArr InvertArray(const IntArr &arr, uint size) const;
	IntArr Merge(const IntArr&, const IntArr &, const T *data, uint data_size) const;
	IntArr MakeOddArr(const T *data, uint data_size, uint &sym, uint alphabet) const;
	IntArr MakeEvenArr(const T *data, const IntArr &arr, uint alphabet) const;
	IntArr Compress(const T *data, uint data_size, uint &sym, uint alphabet) const;
	void Decompress(IntArr &arr, uint data_size) const;
public:
	IntArr CreateSuffix(std::vector<T> &data, uint alphabet = 256) const;
	IntArr CalcLcp(const std::vector<T> &data, const IntArr &arr) const;
};

template<typename T>
void StableCountSort(const IntArr &arr1, IntArr &arr2, const T *data, IntArr &count)
{
	std::fill(count.begin(), count.end(), 0);
	for (uint i : arr1) count[data[i]]++;

	uint s = 0;
	for (uint &c : count) {
		std::swap(s, c);
		s += c;
	}
	for (uint i : arr1) arr2[count[data[i]]++] = i;
}

template<typename T>
IntArr SuffArr<T>::InvertArray(const IntArr &arr, uint size) const
{
	IntArr inv(size);
	for (uint n = 0; n < arr.size(); n++) inv[arr[n]] = n;
	return inv;
}

template<typename T>
IntArr SuffArr<T>::CalcLcp(const std::vector<T> &data, const IntArr &suff) const
{
	IntArr lcp(suff.size());
	IntArr inv = InvertArray(suff, suff.size());
	for (uint n = 0, l = 0; n < suff.size(); n++)
	{
		if (l) l--;
		const uint i = inv[n] + 1;
		if (i < data.size()) {
			const uint m = suff[i];
			while (n + l < data.size() && m + l < data.size() && data[n + l] == data[m + l]) l++;
			lcp[i] = l;
		}
	}
	return lcp;
}


template<typename T>
IntArr SuffArr<T>::Merge(const IntArr &arr1, const IntArr &arr2, const T *data, uint data_size) const
{
	IntArr rank = InvertArray(arr1, data_size + 3);

	auto le2 = [&data, &rank](uint i1, uint i2) {
		int s = data[i1] - data[i2];
		if (!s) s = rank[i1 + 1] - rank[i2 + 1];
		return s < 0;
	};
	auto le3 = [&data, &rank](uint i1, uint i2) {
		int s = data[i1] - data[i2];
		if (!s) s = data[i1 + 1] - data[i2 + 1];
		if (!s) s = rank[i1 + 2] - rank[i2 + 2];
		return s < 0;
	};

	IntArr result_suff;
	result_suff.reserve(data_size);

	uint i1 = 0, i2 = 0;
	while (i1 < arr1.size() && i2 < arr2.size()) {
		bool c = (arr1[i1] % 3 == 1) ? le2(arr1[i1], arr2[i2]) : le3(arr1[i1], arr2[i2]);
		uint i = c ? arr1[i1++] : arr2[i2++];
		if (i < data_size) result_suff.push_back(i);
	}

	while (i1 < arr1.size()) result_suff.push_back(arr1[i1++]);
	while (i2 < arr2.size()) result_suff.push_back(arr2[i2++]);
	assert(result_suff.size() == data_size);
	return result_suff;
}

template<typename T>
IntArr SuffArr<T>::Compress(const T *data, uint data_size, uint &sym, uint alphabet) const
{
	IntArr index1; index1.reserve((data_size + 1) * 2 / 3);
	for (uint i = 0; i <= data_size; i++) if (i % 3) index1.push_back(i);
	assert(index1.size() == (data_size + 1) * 2 / 3);

	IntArr count(alphabet);
	IntArr index2(index1.size());
	StableCountSort(index1, index2, &data[2], count);
	StableCountSort(index2, index1, &data[1], count);
	StableCountSort(index1, index2, &data[0], count);

	auto eq = [&](uint i1, uint i2)->bool {return data[i1] == data[i2] && data[i1 + 1] == data[i2 + 1] && data[i1 + 2] == data[i2 + 2]; };

	sym = 0;
	IntArr result_pack;
	result_pack.reserve(index1.size() + 3);
	result_pack.resize(index1.size());
	uint part_size = (index2.size() + 1) / 2;
	for (uint i = 0; i < index2.size(); i++) {
		if (i && !eq(index2[i - 1], index2[i])) sym++;
		uint j = index2[i];
		j = j - j / 3 - 1;
		j = (j & 1) ? j / 2 + part_size : j / 2;
		result_pack[j] = sym;
	}
	sym++;
	return result_pack;
}

template<typename T>
void SuffArr<T>::Decompress(IntArr &arr, uint data_size) const
{
	uint part_size = (data_size + 2) / 3;
	for (uint &i : arr) i = (i < part_size) ? i * 3 + 1 : (i - part_size) * 3 + 2;
}

template<typename T>
IntArr SuffArr<T>::MakeEvenArr(const T *data, const IntArr &arr, uint alphabet) const
{
	IntArr index; index.reserve((arr.size() + 1) / 2);
	for (uint i : arr) if (i % 3 == 1) index.push_back(i - 1);
	assert(index.size() == (arr.size() + 1) / 2);

	IntArr count(alphabet);
	IntArr result_arr(index.size());
	StableCountSort(index, result_arr, &data[0], count);
	return result_arr;
}

template<typename T>
IntArr SuffArr<T>::MakeOddArr(const T *data, uint data_size, uint &sym_count, uint alphabet) const
{
	IntArr pack = Compress(&data[0], data_size, sym_count, alphabet);
	if (pack.size() == sym_count) return InvertArray(pack, pack.size());
	return SuffArr<uint>().CreateSuffix(pack, sym_count);
}

template<typename T>
IntArr SuffArr<T>::CreateSuffix(std::vector<T> &data, uint alphabet) const
{
	uint data_size = data.size();
	data.push_back(0);
	data.push_back(0);
	data.push_back(0);

	uint sym_count;
	IntArr arr12 = MakeOddArr(&data[0], data_size, sym_count, alphabet);
	Decompress(arr12, data_size);
	IntArr arr0 = MakeEvenArr(&data[0], arr12, alphabet);
	return Merge(arr12, arr0, &data[0], data_size);
}


void test(const char *txt)
{
	// debug output
	for (uint n = 0; txt[n]; n++) {
		putchar(txt[n]);
		if (n > 60) { printf(".."); break; }
	}
	putchar('\n');

	ByteArr data; data.reserve(strlen(txt));
	while (*txt) data.push_back(*txt++);

	// make
	IntArr res = SuffArr<byte>().CreateSuffix(data);
	for (size_t n = 1; n < res.size(); n++)
	{
		uint i1 = res[n - 1];
		uint i2 = res[n];
		if (strcmp((char*)&data[i1], (char*)&data[i2]) > 0) {
			printf("%d %d\n", i1, i2);
			printf("\n\n[%zu] %s\n", strlen((char*)&data[i1]), &data[i1]);
			printf("\n\n[%zu] %s\n", strlen((char*)&data[i2]), &data[i2]);
			assert(0);
		}
	}
}


int main()
{
	// base test
	test("x");
	test("cbbbb");
	test("dedbebdabaabaabaebccadddccadec");
	test("1234567");
	test("abbacab");
	test("misssisippi");
	test("yabbadabbado");
	test("aaabaabaaaababbbabbababbbaaabb");
	test("abaabaabbababbaaaaaabbbbabaabaaabaabaaaababbbabbababbbaaabbbbb");
	test("abaabaabbababb$aaaaaabbbbabaabaaabaabaaaababbbabbababbbaaabbbbb");

	// random test
	char str[1000];
	for (char a = 92; a >= 2; a--) {
		char s = 'a' + a > 'z' ? 33 : 'a';
		for (int p = 0; p < 100000; p++) {
			srand(p); rand();
			size_t l = rand() % 100 + 1;
			size_t m = 0;
			while (m < l) str[m++] = s + rand() % a;
			str[m] = 0;
			printf("%d:%d:", a, p);
			test(str);
		}
	}
	for (uint n = 0; n < 700;) {
		str[n++] = 'a'; str[n] = 0;
		printf("1:%d:", n);
		test(str);
	}

	// big test
	enum { limit = 13000000 };
	char *buff = new char[limit];
	FILE *f = fopen("rfc-index.xml","rb");
	if (f) {
		size_t n = fread(buff, 1, limit - 1, f);
		buff[n] = 0;
		fclose(f);
		printf("file size: %zu\n", strlen(buff));
		test(buff);
		delete[]buff;
	}
	system("pause");
}
